#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <inttypes.h>
#include <getopt.h>

#include "../firmware/smcar-relay-protocol.h"

#define DEFAULT_PORT	"/dev/ttyUSB0"

static int uart_file;

static uint8_t send_buf[ SMCAR_RELAY_UART_RECV_SIZE ] = { SMCAR_RELAY_START_BYTE };
static uint8_t recv_buf[ SMCAR_MAX_ANSWER_SIZE ];

void show_help( char **argv )
{
	printf( "Usage: %s [port] <action>\n", argv[ 0 ] );
	printf( "Where [port] is optional and specified either as -P <device> or --port <device>\n" );
	printf( "\t (default is %s)\n", DEFAULT_PORT );
	printf( "And <action> is one of following:\n" );
	printf( "\t-a, --get-adc\t\tread ADC channels\n" );
	printf( "\t-p, --get-pull\t\tread pull-ups state\n" );
	printf( "\t-s, --set-pull\t\tset pull-ups state (requires an argument)\n" );
	printf( "\t-r, --get-relay\t\tread relays state\n" );
	printf( "\t-R, --set-relay\t\tset relays state (requires an argument)\n" );
	printf( "\t-o, --leds-off\t\tturn off all the LEDs\n" );
	printf( "\t-l, --leds-on\t\tturn on LEDS usage\n" );
	printf( "\t-i, --idle-mode\t\tactivate IDLE mode\n" );
	printf( "\t-S, --get-sonar\t\tread sonar data\n" );
	printf( "\t-h, --help\t\tdisplay this help\n" );
	exit( 1 );
}

void init_port( char *port )
{
	int fl;
        uart_file = open( port, O_RDWR | O_NOCTTY | O_NONBLOCK );
        if ( uart_file == -1 ) {
                perror( "Unable to open communication device" );
		exit( 2 );
	}
	struct termios tios;
        if ( tcgetattr( uart_file, &tios ) < 0 ) {
		perror( "Unable to get parameters, tcgetattr error");
		exit( 2 );
	}
        tios.c_cflag = CS8 | CREAD | HUPCL | CSTOPB;
        tios.c_iflag = IGNBRK | IGNPAR;
        tios.c_oflag = 0;
        tios.c_lflag = 0;
        tios.c_cc[ VMIN ] = 0;
        tios.c_cc[ VTIME ] = 10;
        cfsetospeed( &tios, B57600 );
	cfsetispeed( &tios, B57600 );
        if ( tcsetattr( uart_file, TCSANOW, &tios ) < 0 ) {
                perror( "Unable to set parameters, tcsetattr error" );
		exit( 2 );
	}
        if ( ( fl = fcntl( uart_file, F_GETFL ) ) == -1
	     || fcntl( uart_file, F_SETFL, fl & ~O_NONBLOCK ) < 0 ) {
		perror( "Couldn't reset non-blocking mode on device" );
		exit( 2 );
	}
        tcflush( uart_file, TCIOFLUSH );
}

void send_command( uint8_t answer_size )
{
	int i;
	int cs = 0;
	int cnt, read_bytes=0;
	for ( i = 1; i < SMCAR_RELAY_UART_RECV_SIZE - 1; i++ )
		cs ^= send_buf[ i ];
	send_buf[ SMCAR_RELAY_UART_RECV_SIZE - 1 ] = cs;
	if ( write( uart_file, send_buf, sizeof( send_buf ) ) != sizeof( send_buf ) ) {
		perror( "Couldn't output a command" );
		exit( 3 );
	}
	while ( read_bytes < answer_size ) {
		cnt = read( uart_file, &recv_buf[ read_bytes ], answer_size - read_bytes );
		if ( cnt == -1 ) {
			perror( "Couldn't read a response" );
			exit( 3 );
		}
		read_bytes |= cnt;
	}
	cs = 0;
	for ( i = 0; i < answer_size; i++ )
		cs ^= recv_buf[ i ];
	if ( cs ) {
		fprintf( stderr, "Checksum returned is incorrect\n" );
		exit( 3 );
	}
}

void read_adc()
{
	int i;
	send_buf[ 1 ] = SMCAR_RELAY_GET_ADC;
	send_command( 27 );
	for ( i = 0; i < 24; i++ )
		printf( "adc%d:\t%f\n", i + 1, recv_buf[ i ] / 10. );
	printf( "Ign:\t%f\n", recv_buf[ 24 ] / 10. );
	printf( "Bat:\t%f\n", recv_buf[ 25 ] / 10. );
}

void get_pull()
{
	send_buf[ 1 ] = SMCAR_RELAY_GET_PULLUP;
	send_command( 4 );
	printf( "Pullups state: 0x%.2x 0x%.2x 0x%.2x\n", recv_buf[ 0 ], recv_buf[ 1 ], recv_buf[ 2 ] );
}

void get_relay()
{
	send_buf[ 1 ] = SMCAR_RELAY_GET_RELAY;
	send_command( 4 );
	printf( "Relays state: 0x%.2x 0x%.2x 0x%.2x\n", recv_buf[ 0 ], recv_buf[ 1 ], recv_buf[ 2 ] );
}

void get_sonar()
{
	send_buf[ 1 ] = SMCAR_RELAY_GET_PARK;
	send_command( 3 );
	printf( "Sonar time: %d\n", recv_buf[ 0 ] | recv_buf[ 1 ]<<8 );
}

void activate_idle()
{
	send_buf[ 1 ] = SMCAR_RELAY_ACTIVATE_IDLE;
	send_command( 1 );
}

void leds_off()
{
	send_buf[ 1 ] = SMCAR_RELAY_LEDS_OFF;
	send_command( 1 );
}

void leds_on()
{
	send_buf[ 1 ] = SMCAR_RELAY_LEDS_ON;
	send_command( 1 );
}

void set_pull( int arg1, int arg2, int arg3 )
{
	send_buf[ 1 ] = SMCAR_RELAY_SET_PULLUP;
	send_buf[ 2 ] = (uint8_t)arg1;
	send_buf[ 3 ] = (uint8_t)arg2;
	send_buf[ 4 ] = (uint8_t)arg3;
	send_command( 1 );
}

void set_relay( int arg1, int arg2, int arg3 )
{
	send_buf[ 1 ] = SMCAR_RELAY_SET_RELAY;
	send_buf[ 2 ] = (uint8_t)arg1;
	send_buf[ 3 ] = (uint8_t)arg2;
	send_buf[ 4 ] = (uint8_t)arg3;
	send_command( 1 );
}

int main( int argc, char **argv )
{
	char *port = DEFAULT_PORT;
	int action;
	char *action_arg;
	static struct option long_options[] = {
		{ "port", required_argument, NULL, 'P' },
		{ "get-adc", 0, NULL, 'a' },
		{ "get-pull", 0, NULL, 'p' },
		{ "set-pull", required_argument, NULL, 's' },
		{ "get-relay", 0, NULL, 'r' },
		{ "set-relay", required_argument, NULL, 'R' },
		{ "leds-off", 0, NULL, 'o' },
		{ "leds-on", 0, NULL, 'l' },
		{ "idle-mode", 0, NULL, 'i' },
		{ "get-sonar", 0, NULL, 'S' },
		{ "help", 0, NULL, 'h' },
		{ NULL, 0, NULL, 0 } };
	while ( 1 ) {
		int c;
		c = getopt_long( argc, argv, "P:aps:rR:oliSh", long_options, NULL );
		if ( c == -1 )
			break;
		switch ( c ) {
		case 'P':
			port = optarg;
			break;
		case 'h':
		case '?':
			show_help( argv );
			break;
		default:
			action = c;
			action_arg = optarg;
		}
	}
	if ( optind < argc || argc == 1 )
		show_help( argv );
	init_port( port );
	switch ( action ) {
	case 'a':
		read_adc();
		break;
	case 'p':
		get_pull();
		break;
	case 'i':
		activate_idle();
		break;
	case 'r':
		get_relay();
		break;
	case 's': {
		int arg1, arg2, arg3;
		sscanf( action_arg, "%i%i%i", &arg1, &arg2, &arg3 );
		set_pull( arg1, arg2, arg3 );
		break;
	}
	case 'R': {
		int arg1, arg2, arg3;
		sscanf( action_arg, "%i%i%i", &arg1, &arg2, &arg3 );
		set_relay( arg1, arg2, arg3 );
		break;
	}
	case 'o':
		leds_off();
		break;
	case 'l':
		leds_on();
		break;
	case 'S':
		get_sonar();
		break;
	}
	close( uart_file );
	return 0;
}
